"""implement api to access google sheet
"""
import logging
from typing import Optional

from googleapiclient.discovery import Resource

VALID_DIMENSION = {"COLUMNS", "ROWS"}


class Sheets:
    """provide api to operate google spreadsheet. An authenticated google api client is needed.
    """
    def __init__(self, service: Resource):
        self._service = service.spreadsheets()

    def download(self, id: str, range: str, dimension: str="ROWS") -> list:
        """download target sheet range by specified dimension. All entries will be considered as strings.

        :param id: id of the target spreadsheet.
        :param range: range in the target spreadsheet. for example, 'sheet!A1:D'. this means selecting from tab "sheet"
          and download column A to D and rows from 1 to the last row with non-empty values.
        :param dimension: "ROW" or "COLUMNS". If "ROWS", each entry in the output list would be one row in the
          spreadsheet. If "COLUMNS", each entry in the output list would be one column in the spreadsheet.
        :return: content of target sheet range in a list of lists.
        """
        if dimension not in VALID_DIMENSION:
            raise ValueError(f"{dimension} is not a valid dimension. expecting {VALID_DIMENSION}.")

        result = self._service.values().get(spreadsheetId=id,
                                            range=range,
                                            majorDimension=dimension).execute()
        values = result.get('values', [])
        return values

    def upload(self, values: list, id: str, range: str) -> None:
        """upload a list of lists to target sheet range.

        :param values: a list of lists of objects that can be converted to str.
        :param id: id of the target spreadsheet
        :param range: range in the target spreadsheet. for example, 'sheet!A1:D'. this means selecting from tab "sheet"
          and download column A to D and rows from 1 to the last row with non-empty values.
        :return: None
        """
        self.clear(id=id, range=range)
        body = {"values": values}
        logging.info(f"Updating sheet '{id}' range '{range}'")
        request = self._service.values().update(spreadsheetId=id,
                                                range=range,
                                                valueInputOption="RAW",
                                                body=body)
        result = request.execute()
        msg = f"{result.get('updatedRange')} has been updated ({result.get('updatedRows')} rows " \
              f"and {result.get('updatedColumns')} columns)"
        logging.info(msg)

    def clear(self, id: str, range: str):
        """remove content in the target sheet range.

        :param id: id of the target spreadsheet
        :param range: range in the target spreadsheet.  for example, 'sheet!A1:D'. this means selecting from tab "sheet"
          and download column A to D and rows from 1 to the last row with non-empty values.
        :return: None
        """
        self._service.values().clear(spreadsheetId=id, range=range, body={}).execute()

    def read_sheet(self, id: str, range: str, header=True, dtypes: Optional[dict]=None, columns: Optional[list]=None):
        """download the target sheet range into a pandas datafrme. this method will fail if pandas cannot be imported.

        :param id: id of the target spreadsheet
        :param range: range in the target spreadsheet.  for example, 'sheet!A1:D'. this means selecting from tab "sheet"
          and download column A to D and rows from 1 to the last row with non-empty values.
        :param header: whether first row is used as column names in the output dataframe.
        :param dtypes: a mapping from column name to the type. if not None, type conversions will be applied to columns
          requested in the dictionary.
        :param columns: a list of column names. If not None and `header` is False, this will be used as columns of the
          output dataframe
        :return: a pandas dataframe containing target spreadsheet values.
        """
        try:
            import pandas as pd
        except ModuleNotFoundError as e:
            logging.critical("read_sheet() requires pandas.")
            raise e

        if dtypes is not None and not isinstance(dtypes, dict):
            raise TypeError(f"dtypes must be dictionary. got {type(dtypes)}")

        values = self.download(id=id, range=range)
        if values == []:
            return pd.DataFrame()

        columns = values.pop(0) if header else columns
        df = pd.DataFrame(values, columns=columns)

        if dtypes is not None:
            for col, type in dtypes.items():
                df[col] = df[col].astype(type)

        return df

    def to_sheet(self, df, id: str, range: str):
        """upload pandas dataframe to target sheet range. the number of columns must fit the range. more columns or
        fewer columns will both raise exception.

        :param df: pandas dataframe to be uploaded
        :type df: pandas.DataFrame
        :param id: id of the target spreadsheet
        :param range: range in the target spreadsheet.  for example, 'sheet!A1:D'. this means selecting from tab "sheet"
          and download column A to D and rows from 1 to the last row with non-empty values.
        :return: None
        """
        values = df.fillna('').values.tolist()
        values.insert(0, list(df.columns))
        self.upload(values, id=id, range=range)

    def create_spreadsheet(self, name: str) -> str:
        """create a spreadsheet with requested name.

        :param name: name of the created sheet.
        :return: id of the spreadsheet
        """
        file_metadata = {
            "properties": {"title": name}
        }
        response = self._service.create(body=file_metadata, fields="spreadsheetId").execute()
        return response.get("spreadsheetId")

    def batch_update(self, id: str, body: dict):
        """low level api used to submit a json body to make changes to the specified spreadsheet.

        :param id: id of the target spreadsheet
        :param body: request json
        :return: response from batch upate
        """
        response = self._service.batchUpdate(spreadsheetId=id, body=body).execute()
        return response

    def create_sheet(self, id: str, title: str):
        """create a new sheet with given name in the specified spreadsheet.

        :param id: id of the spreadsheet
        :param title: title of the new sheet
        :return: a dictionary containing information about created sheet, such as sheet id, title, index.
        """
        request = {"requests": [{"addSheet": {"properties": {"title": title}}}]}
        response = self.batch_update(id=id, body=request)
        info = response.get("replies")[0]["addSheet"]["properties"]
        return info

    def delete_sheet(self, id: str, sheet_id: int):
        """delete the specified sheet in the target spreadsheet. You can find sheet_id from URL when you select the
        sheet in the spreadsheet after "gid="

        :param id: id of spreadsheet.
        :param sheet_id: id of sheet
        :return: None
        """
        request = {"requests": [{"deleteSheet": {"sheetId": sheet_id}}]}
        self.batch_update(id=id, body=request)

    def rename_sheet(self, id: str, sheet_id: int, title: str):
        """rename a sheet in target spreadsheet to the new title.

        :param id: id of the target spreadsheet
        :param sheet_id: id of the sheet
        :param title: new title of the sheet
        :return: None
        """
        request = {"requests": [
            {"updateSheetProperties": {"properties": {"sheetId": sheet_id, "title": title}, "fields": "title"}}
        ]
        }
        self.batch_update(id=id, body=request)



"""implement api to access google drive
"""
import logging
from pathlib import PosixPath, Path
from typing import Union, Optional, List
import re

from googleapiclient.discovery import Resource
from googleapiclient.http import MediaIoBaseDownload, MediaFileUpload


class Drive:

    def __init__(self, service: Resource):
        self._service = service

    def download(self, id: str, to_file: Union[str, PosixPath]):
        """download the google drive file with the requested id to target local file.

        :param id: id of the google drive file
        :param to_file: local file path
        :return: None
        """
        request = self._service.files().get_media(fileId=id)
        with open(to_file, 'wb') as fh:
            downloader = MediaIoBaseDownload(fh, request)
            done = False
            while not done:
                status, done = downloader.next_chunk()
                logging.info(f"Download {status.progress()*100}%")

    def upload(self, from_file: Union[str, PosixPath], name: Optional[str]=None, mimetype: Optional[str]=None,
               parent_ids: Optional[List[str]]=None) -> str:
        """upload local file to gdrive.

        :param from_file: path to local file.
        :param name: name of google drive file. If None, the name of local file will be used.
        :param mimetype: Mime-type of the file. If None then a mime-type will be guessed from the file extension.
        :param parent_ids: list of ids for the folder you want to upload the file to. If None, it will be uploaded to
          root of Google drive.
        :return: id of the uploaded file
        """
        file_metadata = {'name': name if name is not None else Path(from_file).name}

        if parent_ids is not None:
            if not isinstance(parent_ids, list):
                raise TypeError(f"parent_ids must be a list. got {type(parent_ids)}")

            if len(parent_ids) == 0:
                raise ValueError(f"parent_ids cannot be empty")

            file_metadata["parents"] = parent_ids

        media = MediaFileUpload(str(from_file),
                                mimetype=mimetype,
                                resumable=True)

        file = self._service.files().create(body=file_metadata,
                                            media_body=media,
                                            fields='id').execute()
        return file.get("id")

    def update(self, id: str, from_file: Union[str, PosixPath]):
        """update the Google drive with local file.

        :param id: id of the Google drive file to be updated
        :param from_file: path to local file.
        :return: None
        """
        media = MediaFileUpload(str(from_file),
                                resumable=True)

        self._service.files().update(body=dict(), fileId=id, media_body=media).execute()

    def get_id(self, name: str, parent_id: Optional[str]=None):
        """get the id of the file with specified name. if more than one file are found, an error will be raised.

        :param name: name of the file to be searched.
        :param parent_id: id of the folder to limit the search. If None, the full Google drive will be searched.
        :return: the id of the file if found. Or None if no such name is found.
        """
        q = f"name = '{name}' and trashed = false"
        if parent_id is not None:
            q += f" and '{parent_id}' in parents"

        response = self._service.files().list(pageSize=10,
                                              fields=self._get_fields_query_string(),
                                              q=q).execute()

        item = response.get('files', None)
        if item is None:
            return None

        if len(item) > 1:
            raise RuntimeError(f"More than one file is found. Please rename the file with a unique string.")

        return item[0]['id']

    def find(self, name_contains: Optional[str]=None, name_not_contains: Optional[str]=None,
             parent_id: Optional[str]=None) -> list:
        """find all files whose name contain specified string and do not contain specified string. Note that Google
        API has unexpected behavior when searching for strings in name. It is can only search first 26 character. In
        addition, it seems to search from first alphabetic character and Assume there are the following files:
        'positive_a', 'positive_b', 'a', '_a', 'ba'

        :example:

        >>> self.find(name_contains='a')  # this finds only 'a' and '_a', not 'positive_a' or 'ba'

        :param name_contains: a string contained in the name
        :param name_not_contains: a string that is not contained in the name
        :param parent_id: parent folder id
        :return: a list of dictionaries containing id and name of found files.
        """
        if name_contains is None and name_not_contains is None:
            raise ValueError("name_contains and name_not_contains cannot both be None")

        q_name_contains = ""
        q_name_not_contains = ""
        if name_contains is not None:
            q_name_contains = f"and name contains '{name_contains}'"
        if name_not_contains is not None:
            q_name_not_contains = f"and not name contains '{name_not_contains}'"
        q = f"trashed = false {q_name_contains} {q_name_not_contains}"

        if parent_id is not None:
            q += f" and '{parent_id}' in parents"
        response = self._service.files().list(pageSize=100,
                                              fields=self._get_fields_query_string(),
                                              q=q).execute()
        item = response.get('files', [])
        return item

    def list(self, id: str, regex: str=None, recursive: bool=False, depth: int=3) -> list:
        """list the content of the folder by the given id.

        :param id: id of the folder to be listed.
        :param regex: an regular expression used to filter returned file and folders.
        :param recursive: if True, children of the folder will also be listed.
        :param depth: number of recursion if recursive is True. This is to prevent cyclic nesting or deep nested folders.
        :return: a list of dictionaries containing id, name of the object contained in the target folder and list of
          parent ids.
        """
        q = f"'{id}' in parents and trashed = false"
        result = []
        page_token = ""  # place holder to start the loop
        while page_token is not None:
            response = self._service.files().list(q=q,
                                                  spaces='drive',
                                                  fields=self._get_fields_query_string(["id", "name", "parents"]),
                                                  pageToken=page_token).execute()
            result.extend(response.get("files", []))
            page_token = response.get("nextPageToken", None)

        if recursive and depth > 0:
            for file in result:
                children_id = file["id"]
                result.extend(self.list(id=children_id, recursive=True, depth=depth-1))

        if regex is not None:
            pattern = re.compile(regex)
            result = [f for f in result if pattern.match(f["name"])]

        return result

    def delete(self, id: str, recursive: bool=False):
        """delete target file from google drive
        TODO: implement recursive delete

        :param id: id of target object.
        :param recursive: if True and target id represents a folder, remove all nested files and folders.
        :return: None
        """
        self._service.files().delete(fileId=id).execute()

    def create_folder(self, name: str, parent_ids: Optional[list]=None) -> str:
        """create a folder on google drive by the given name.

        :param name: name of the folder to be created.
        :param parent_ids: list of ids where you want to create your folder in.
        :return: id of the created folder.
        """
        file_metadata = {
            'name': name,
            'mimeType': 'application/vnd.google-apps.folder'
        }
        if parent_ids is not None:
            if not isinstance(parent_ids, list):
                raise TypeError(f"parent_ids must be a list. got {type(parent_ids)}")

            if len(parent_ids) == 0:
                raise ValueError(f"parent_ids cannot be empty")

            file_metadata["parents"] = parent_ids

        folder = self._service.files().create(body=file_metadata, fields='id').execute()
        return folder.get("id")

    def share(self, id: str, emails: List[str], role: str= "reader", notify=True):  # pragma: no cover
        """modify the permission of the target object and share with the provided emails.

        :param id: id of target object.
        :param emails: list of emails to be shared with.
        :param role: type of permission. accepted values are: 'owner', 'organizer', 'fileOrganzier', 'writer',
          'commenter' and 'reader'.
        :param notify: Whether notifying emails about the sharing.
        :return: name of the object shared.
        """
        call_back = None
        batch = self._service.new_batch_http_request(callback=call_back)

        for email in emails:
            user_permission = {
                "type": "user",
                "role": role,
                "emailAddress": email
            }
            batch.add(self._service.permissions().create(
                fileId=id,
                body=user_permission,
                fields='id',
                sendNotification=notify
            ))
        batch.execute()
        return self.get_name(id)

    def _get_fields_query_string(self, fields: Optional[list]=None) -> str:
        """create a string used to query gdrive object and return requested fields.

        :param fields: list of fields to be returned in query.
        :return: a string used to query gdrive. only usable in `fields` arguments in list()
        """
        if fields is None:
            fields = ["id", "name"]

        if not isinstance(fields, list):
            raise TypeError(f"fields must be a list. got {type(fields)}")

        if len(fields) == 0:
            raise ValueError("fields cannot be empty")

        return f"nextPageToken, files({','.join(fields)})"

    def get_name(self, id: str) -> str:
        """get the name of the Google drive object.

        :param id: id of the target Google drive object
        :return: name of the object
        """
        file = self._service.files().get(fileId=id).execute()
        return file['name']

    def copy(self, id: str, name: str, parent_ids: Optional[list]=None) -> str:
        """copy target file and give the new file specified name. return the id of the created file.

        :param id: target file to be copied.
        :param name: name of the new file.
        :param parent_ids: ids of folders to place the new file in.
        :return: id of the created new file.
        """
        request = {"name": name}
        if parent_ids is not None:
            request["parents"] = parent_ids
        file = self._service.files().copy(fileId=id, body=request, fields='id').execute()
        return file.get("id")
